-- |
-- Module      :  Languages.UniquenessPeriods.Vector.ConstraintsG.Encoded
-- Copyright   :  (c) OleksandrZhabenko 2020
-- License     :  MIT
-- Stability   :  Experimental
-- Maintainer  :  olexandr543@yahoo.com
--
-- Provides a way to encode the needed constraint with possibly less symbols.
--

{-# LANGUAGE FlexibleInstances #-}

module Languages.UniquenessPeriods.Vector.ConstraintsG.Encoded where

import Data.Monoid (mappend)
import Text.Read (readMaybe)
import Data.Maybe
import qualified Data.Vector as VB
import Data.List (intercalate,sort,nub)
import Languages.UniquenessPeriods.Vector.ConstraintsG

data EncodedContraints a b = E a | T a a a | SA a b | SB a b | F a a deriving (Eq, Ord)

-- | Inspired by the: https://hackage.haskell.org/package/base-4.14.0.0/docs/Data-Maybe.html
readMaybeEC :: String -> Maybe EncodedCnstrs
readMaybeEC xs
 | null xs = Just (E 0)
 | otherwise =
     let h = take 1 xs
         ts = nub . sort . filter (\x -> x >= '0' && x <= '6') . drop 1 $ xs in
      case h of
       "E" -> Just (E (fromMaybe 0 (readMaybe (take 1 . tail $ xs)::Maybe Int)))
       "F" -> let (y,z) = (readMaybe (take 1 ts)::Maybe Int, readMaybe (take 1 . drop 1 $ ts)) in
         case (y,z) of
          (Nothing,r2) -> Nothing
          (r1,Nothing) -> Nothing
          ~(Just x1, Just x2) -> Just (F x1 x2)
       "T" -> let (y,z,u) = (readMaybe (take 1 ts)::Maybe Int, readMaybe (take 1 . drop 1 $ ts)::Maybe Int, readMaybe (take 1 . drop 2 $ ts)::Maybe Int) in
         case (y,z,u) of
          (Nothing,r2,r3) -> Nothing
          (r1,Nothing,r3) -> Nothing
          (r1,r2,Nothing) -> Nothing
          ~(Just x1, Just x2, Just x3) -> Just (T x1 x2 x3)
       "A" -> let y = readMaybe (take 1 ts)::Maybe Int in
               if isJust y then
                let y0 = fromJust y
                    zs = filter (/= y0) . catMaybes . map (\t -> readMaybe [t]::Maybe Int) . drop 1 $ ts in
                     case zs of
                       [] -> Nothing
                       x2 -> Just (SA y0 (VB.fromList x2))
               else Nothing
       "B" -> let y = readMaybe (take 1 ts)::Maybe Int in
               if isJust y then
                let y0 = fromJust y
                    zs = filter (/= y0) . catMaybes . map (\t -> readMaybe [t]::Maybe Int) . drop 1 $ ts in
                     case zs of
                       [] -> Nothing
                       x2 -> Just (SB y0 (VB.fromList x2))
               else Nothing
       _   -> Nothing

type EncodedCnstrs = EncodedContraints Int (VB.Vector Int)

-- | Must be applied to the correct vector of permutation indeces. Otherwise, it gives runtime error (exception).
decodeConstraint1 :: EncodedCnstrs -> VB.Vector (VB.Vector Int) -> VB.Vector (VB.Vector Int)
decodeConstraint1 (E _) = id
decodeConstraint1 (T i j k) = unsafeTriples i j k
decodeConstraint1 (SA i v) = unsafeSeveralA i v
decodeConstraint1 (SB i v) = unsafeSeveralB i v
decodeConstraint1 (F i j) = filterOrderIJ i j

decodeLConstraints :: [EncodedCnstrs] -> VB.Vector (VB.Vector Int) -> VB.Vector (VB.Vector Int)
decodeLConstraints (x:xs) = decodeLConstraints' ys . decodeConstraint1 y
  where y = minimum (x:xs)
        ys = filter (/= y) . g $ (x:xs)
        g ((E _):xs) = g xs
        g (x:xs) = x : g xs
        g _ = []
        decodeLConstraints' (z:zs) = decodeLConstraints' zs . decodeConstraint1 z
        decodeLConstraints' _ = id
decodeLConstraints _ = id

isE :: EncodedCnstrs -> Bool
isE (E _) = True
isE _ = False

isF :: EncodedCnstrs -> Bool
isF (F _ _) = True
isF _ = False

isT :: EncodedCnstrs -> Bool
isT (T _ _ _) = True
isT _ = False

isSA :: EncodedCnstrs -> Bool
isSA (SA _ _) = True
isSA _ = False

isSB :: EncodedCnstrs -> Bool
isSB (SB _ _) = True
isSB _ = False

getIEl :: EncodedCnstrs -> Int
getIEl (E i) = i
getIEl (T i _ _) = i
getIEl (SA i _) = i
getIEl (SB i _) = i
getIEl (F i _) = i

setIEl :: Int -> EncodedCnstrs -> EncodedCnstrs
setIEl i (E _) = E i
setIEl i (T _ j k) = T i j k
setIEl i (SA _ v) = SA i v
setIEl i (SB _ v) = SB i v
setIEl i (F _ j) = F i j


